package org.vaadin.imagefilter;

import java.awt.Graphics2D;
import java.awt.image.BufferedImage;

import org.vaadin.imagefilter.utils.ImageUtils;

/**
 * Abstract image filter operation to draw. This {@link Image.Operation}
 * implementation is meant to be subclassed. It does nt do enything but copy the
 * original image. You should override either
 * {@link #draw(Graphics2D, BufferedImage)} or {@link #apply(BufferedImage)}
 * method.
 *
 * @author Sami Ekblad
 *
 */
public abstract class AbstractDrawOperation implements Image.Operation {

    private static final String DEFAULT_NAME = "Custom Draw Operation";

    /**
     * Create new filter operation.
     *
     * @param filterOp
     */
    public AbstractDrawOperation() {
    }

    /**
     * Apply the operation to the given image. This should always create a new
     * destination image.
     *
     *Default implementation creates a new image and invokes
     * {@link #draw(Graphics2D, BufferedImage)} method.
     *
     * @param image
     *            The image to be processed
     * @return A new image instance, after applying the operation to image.
     *
     */
    public BufferedImage apply(BufferedImage image) {
        BufferedImage dst = ImageUtils.createBufferedImage(image.getWidth(),
                image.getHeight());
        Graphics2D g = dst.createGraphics();
        ImageUtils.applyDefaultRenderingHints(g);
        try {
            draw(g, image);
        } finally {
            if (g != null) {
                g.dispose();
            }
        }
        return dst;
    }

    public abstract void draw(Graphics2D g, BufferedImage srcImage);

    @Override
    public String toString() {
        return "" + getName();
    }

    /**
     * Get name of the operation.
     *
     * @return
     */
    public String getName() {
        return DEFAULT_NAME;
    }

}