package org.vaadin.imagefilter;

import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.Serializable;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import com.jhlabs.image.AbstractBufferedImageOp;

/**
 * Image filter operation.
 *
 * This image operation simply wraps a {@link AbstractBufferedImageOp} instance.
 *
 * @author Sami Ekblad
 *
 */
public class FilterOperation implements Image.Operation, Serializable {

    private static final long serialVersionUID = 848361488313774794L;

    private AbstractBufferedImageOp filter;

    /**
     * Get filter operation.
     *
     * @return
     */
    public AbstractBufferedImageOp getFilter() {
        return filter;
    }

    /**
     * Create new filter operation.
     *
     * @param filterOp
     */
    FilterOperation(AbstractBufferedImageOp filterOp) {
        filter = filterOp;
    }

    public BufferedImage apply(BufferedImage image) {
        return filter.filter(image, null);
    }

    @Override
    public String toString() {
        return "FilterOperation: " + getName();
    }

    /**
     * Get name of the filter
     *
     * @return
     */
    public String getName() {
        String name = filter.getClass().getSimpleName();
        if (name.endsWith("Filter")) {
            name = name.substring(0, name.length() - 6);
        }
        return name;
    }

    private static final String FILTER_PACKAGE_1 = "org/vaadin/imagefilter/filters";
    private static final String FILTER_PACKAGE_2 = "com/jhlabs/image";

    private static final String[] BLACKLIST_A = new String[] {
            "AbstractBufferedImageOp", "Binary", "Compound", "Iterated",
            "Pointr", "Sky", "Transfer", "Transform", "Transition",
            "WholeImage", "ImageCombining", "Mutatable" };
    private static final Collection<String> BLACKLIST = Arrays
            .asList(BLACKLIST_A);

    @SuppressWarnings("unchecked")
    public static void main(String[] args) {

        try {

            List<String> l = listOperations();
            removeBlacklisted(l, BLACKLIST);
            for (String f : l) {
                String name = getShortName((Class<AbstractBufferedImageOp>) Class
                        .forName(f));
                System.out.println("public static final String "
                        + name.toUpperCase() + " = \"" + name + "\";");
            }

            System.out
                    .println("public static final Set<String> ALL_OPERATIONS;");
            System.out
                    .println("public static final Map<String, Class<? extends BufferedImageOp>> ALL_OPERATIONS_MAP;");
            System.out.println("static {");
            System.out
                    .println("LinkedHashMap<String, Class<? extends BufferedImageOp>> m = new LinkedHashMap<String, Class<? extends BufferedImageOp>>();");
            int i = 0;
            for (String f : l) {
                String name = getShortName((Class<AbstractBufferedImageOp>) Class
                        .forName(f));
                System.out.println("m.put(" + name.toUpperCase() + "," + f
                        + ".class);");
                i++;
            }
            System.err.println(i);
            System.out
                    .println("ALL_OPERATIONS_MAP = java.util.Collections.unmodifiableMap(m);");
            System.out
                    .println("ALL_OPERATIONS = java.util.Collections.unmodifiableSet(m.keySet());");
            System.out.println("}");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @SuppressWarnings("unchecked")
    private static void removeBlacklisted(List<String> l,
            Collection<String> blacklist) {
        ArrayList bl = new ArrayList();
        for (String b : blacklist) {
            bl.add(FILTER_PACKAGE_2.replace('/', '.') + "." + b + "Filter");
        }
        l.removeAll(bl);
    }

    public static String getShortName(Class<AbstractBufferedImageOp> f) {
        String n = f.getSimpleName();
        return n.substring(0, n.length() - 6);
    }

    public static List<String> listOperations() {
        List<String> l = listOperations(FILTER_PACKAGE_1);
        l.addAll(listOperations(FILTER_PACKAGE_2));
        Collections.sort(l, new Comparator<String>() {

            public int compare(String s0, String s1) {
                s0 = s0.substring(s0.lastIndexOf('.'));
                s1 = s1.substring(s1.lastIndexOf('.'));
                return s0.compareTo(s1);
            }
        });
        return l;
    }

    public static List<String> listOperations(String filterPkg) {

        List<String> l = new ArrayList<String>();
        ClassLoader cld = FilterOperation.class.getClassLoader();
        try {
            Enumeration<URL> resources = cld.getResources(filterPkg);
            while (resources.hasMoreElements()) {
                URL res = resources.nextElement();
                if (res.getProtocol().equalsIgnoreCase("jar")) {
                    JarURLConnection conn = (JarURLConnection) res
                            .openConnection();
                    JarFile jar = conn.getJarFile();
                    for (JarEntry e : Collections.list(jar.entries())) {

                        if (e.getName().startsWith(filterPkg)
                                && e.getName().endsWith("Filter.class")) {
                            String fname = e.getName().replace('/', '.')
                                    .substring(0, e.getName().length() - 6);
                            l.add(fname);
                        }
                    }
                } else if (res.getProtocol().equalsIgnoreCase("file")) {
                    String dirName = res.toExternalForm().substring(5);
                    File d = new File(dirName);
                    String[] files = d.list(new FilenameFilter() {

                        public boolean accept(File d, String n) {
                            return n.endsWith("Filter.class");
                        }
                    });
                    for (int i = 0; i < files.length; i++) {
                        String f = files[i];
                        l.add(filterPkg.replace('/', '.') + "."
                                + f.substring(0, f.length() - 6));
                    }
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

        return l;
    }

    private static Class<?> getFilterClass(String name) {
        return ALL_OPERATIONS_MAP.get(name);
    }

    private static AbstractBufferedImageOp createBufferedImageOp(String name) {
        Class<?> cls = getFilterClass(name);

        // Stop here if not found
        if (cls == null) {
            return null;
        }

        // Instantiate
        try {
            return (AbstractBufferedImageOp) cls.newInstance();
        } catch (InstantiationException e) {
            System.err.println("Failed to instantiate filter " + name);
        } catch (IllegalAccessException e) {
            System.err.println("Failed to instantiate filter " + name);
        }
        return null;
    }

    /**
     * Get FilterOperation based on name. Creates a new FilterOperation
     * instance.
     *
     * @param filterName
     *            one of the filters specified in {@link FilterOperation} class.
     * @return
     */
    public static FilterOperation getByName(String filterName) {
        AbstractBufferedImageOp filter = createBufferedImageOp(filterName);
        return filter != null ? new FilterOperation(filter) : null;
    }

    public static final String APPLYMASK = "ApplyMask";
    public static final String AVERAGE = "Average";
    public static final String BICUBICSCALE = "BicubicScale";
    public static final String BLOCK = "Block";
    public static final String BLUR = "Blur";
    public static final String BORDER = "Border";
    public static final String BOXBLUR = "BoxBlur";
    public static final String BRUSHEDMETAL = "BrushedMetal";
    public static final String BUMP = "Bump";
    public static final String CAUSTICS = "Caustics";
    public static final String CELLULAR = "Cellular";
    public static final String CHANNELMIX = "ChannelMix";
    public static final String CHECK = "Check";
    public static final String CHROME = "Chrome";
    public static final String CIRCLE = "Circle";
    public static final String COMPOSITE = "Composite";
    public static final String CONTOUR = "Contour";
    public static final String CONTRAST = "Contrast";
    public static final String CONVOLVE = "Convolve";
    public static final String CROP = "Crop";
    public static final String CRYSTALLIZE = "Crystallize";
    public static final String CURL = "Curl";
    public static final String DESPECKLE = "Despeckle";
    public static final String DIFFUSE = "Diffuse";
    public static final String DIFFUSION = "Diffusion";
    public static final String DILATE = "Dilate";
    public static final String DISPLACE = "Displace";
    public static final String DISSOLVE = "Dissolve";
    public static final String DITHER = "Dither";
    public static final String DROPSHADOW = "DropShadow";
    public static final String EDGE = "Edge";
    public static final String EMBOSS = "Emboss";
    public static final String EQUALIZE = "Equalize";
    public static final String ERODE = "Erode";
    public static final String EXPOSURE = "Exposure";
    public static final String FBM = "FBM";
    public static final String FADE = "Fade";
    public static final String FEEDBACK = "Feedback";
    public static final String FIELDWARP = "FieldWarp";
    public static final String FILL = "Fill";
    public static final String FITINTO = "FitInto";
    public static final String FLARE = "Flare";
    public static final String FLIP = "Flip";
    public static final String FLUSH3D = "Flush3D";
    public static final String FOURCOLOR = "FourColor";
    public static final String GAIN = "Gain";
    public static final String GAMMA = "Gamma";
    public static final String GAUSSIAN = "Gaussian";
    public static final String GLINT = "Glint";
    public static final String GLOW = "Glow";
    public static final String GRADIENT = "Gradient";
    public static final String GRADIENTWIPE = "GradientWipe";
    public static final String GRAY = "Gray";
    public static final String GRAYSCALE = "Grayscale";
    public static final String HSBADJUST = "HSBAdjust";
    public static final String HALFTONE = "Halftone";
    public static final String INTERPOLATE = "Interpolate";
    public static final String INVERTALPHA = "InvertAlpha";
    public static final String INVERT = "Invert";
    public static final String JAVALNF = "JavaLnF";
    public static final String KALEIDOSCOPE = "Kaleidoscope";
    public static final String KEY = "Key";
    public static final String LENSBLUR = "LensBlur";
    public static final String LEVELS = "Levels";
    public static final String LIFE = "Life";
    public static final String LIGHT = "Light";
    public static final String LOOKUP = "Lookup";
    public static final String MAPCOLORS = "MapColors";
    public static final String MAP = "Map";
    public static final String MARBLE = "Marble";
    public static final String MARBLETEX = "MarbleTex";
    public static final String MASK = "Mask";
    public static final String MAXIMUM = "Maximum";
    public static final String MEDIAN = "Median";
    public static final String MINIMUM = "Minimum";
    public static final String MIRROR = "Mirror";
    public static final String MOTIONBLUR = "MotionBlur";
    public static final String NOISE = "Noise";
    public static final String OFFSET = "Offset";
    public static final String OIL = "Oil";
    public static final String OPACITY = "Opacity";
    public static final String OUTLINE = "Outline";
    public static final String PERSPECTIVE = "Perspective";
    public static final String PINCH = "Pinch";
    public static final String PLASMA = "Plasma";
    public static final String POINT = "Point";
    public static final String POINTILLIZE = "Pointillize";
    public static final String POLAR = "Polar";
    public static final String POSTERIZE = "Posterize";
    public static final String QUANTIZE = "Quantize";
    public static final String QUILT = "Quilt";
    public static final String RGBADJUST = "RGBAdjust";
    public static final String RAYS = "Rays";
    public static final String REDUCENOISE = "ReduceNoise";
    public static final String RENDERTEXT = "RenderText";
    public static final String RESCALE = "Rescale";
    public static final String RIPPLE = "Ripple";
    public static final String ROTATE = "Rotate";
    public static final String ROUNDEDCORNERS = "RoundedCorners";
    public static final String SCALE = "Scale";
    public static final String SHADE = "Shade";
    public static final String SHADOW = "Shadow";
    public static final String SHAPE = "Shape";
    public static final String SHARPEN = "Sharpen";
    public static final String SHATTER = "Shatter";
    public static final String SHEAR = "Shear";
    public static final String SKELETON = "Skeleton";
    public static final String SMARTBLUR = "SmartBlur";
    public static final String SMEAR = "Smear";
    public static final String SOLARIZE = "Solarize";
    public static final String SPARKLE = "Sparkle";
    public static final String SPHERE = "Sphere";
    public static final String STAMP = "Stamp";
    public static final String SWIM = "Swim";
    public static final String TEXTURE = "Texture";
    public static final String THRESHOLD = "Threshold";
    public static final String TILEIMAGE = "TileImage";
    public static final String TWIRL = "Twirl";
    public static final String UNSHARP = "Unsharp";
    public static final String VARIABLEBLUR = "VariableBlur";
    public static final String WARP = "Warp";
    public static final String WATER = "Water";
    public static final String WEAVE = "Weave";
    public static final String WOOD = "Wood";
    public static final Set<String> ALL_OPERATIONS;
    public static final Map<String, Class<? extends BufferedImageOp>> ALL_OPERATIONS_MAP;
    static {
        LinkedHashMap<String, Class<? extends BufferedImageOp>> m = new LinkedHashMap<String, Class<? extends BufferedImageOp>>();
        m.put(APPLYMASK, com.jhlabs.image.ApplyMaskFilter.class);
        m.put(AVERAGE, com.jhlabs.image.AverageFilter.class);
        m.put(BICUBICSCALE, com.jhlabs.image.BicubicScaleFilter.class);
        m.put(BLOCK, com.jhlabs.image.BlockFilter.class);
        m.put(BLUR, com.jhlabs.image.BlurFilter.class);
        m.put(BORDER, com.jhlabs.image.BorderFilter.class);
        m.put(BOXBLUR, com.jhlabs.image.BoxBlurFilter.class);
        m.put(BRUSHEDMETAL, com.jhlabs.image.BrushedMetalFilter.class);
        m.put(BUMP, com.jhlabs.image.BumpFilter.class);
        m.put(CAUSTICS, com.jhlabs.image.CausticsFilter.class);
        m.put(CELLULAR, com.jhlabs.image.CellularFilter.class);
        m.put(CHANNELMIX, com.jhlabs.image.ChannelMixFilter.class);
        m.put(CHECK, com.jhlabs.image.CheckFilter.class);
        m.put(CHROME, com.jhlabs.image.ChromeFilter.class);
        m.put(CIRCLE, com.jhlabs.image.CircleFilter.class);
        m.put(COMPOSITE, com.jhlabs.image.CompositeFilter.class);
        m.put(CONTOUR, com.jhlabs.image.ContourFilter.class);
        m.put(CONTRAST, com.jhlabs.image.ContrastFilter.class);
        m.put(CONVOLVE, com.jhlabs.image.ConvolveFilter.class);
        m.put(CROP, com.jhlabs.image.CropFilter.class);
        m.put(CRYSTALLIZE, com.jhlabs.image.CrystallizeFilter.class);
        m.put(CURL, com.jhlabs.image.CurlFilter.class);
        m.put(DESPECKLE, com.jhlabs.image.DespeckleFilter.class);
        m.put(DIFFUSE, com.jhlabs.image.DiffuseFilter.class);
        m.put(DIFFUSION, com.jhlabs.image.DiffusionFilter.class);
        m.put(DILATE, com.jhlabs.image.DilateFilter.class);
        m.put(DISPLACE, com.jhlabs.image.DisplaceFilter.class);
        m.put(DISSOLVE, com.jhlabs.image.DissolveFilter.class);
        m.put(DITHER, com.jhlabs.image.DitherFilter.class);
        m
                .put(DROPSHADOW,
                        org.vaadin.imagefilter.filters.DropShadowFilter.class);
        m.put(EDGE, com.jhlabs.image.EdgeFilter.class);
        m.put(EMBOSS, com.jhlabs.image.EmbossFilter.class);
        m.put(EQUALIZE, com.jhlabs.image.EqualizeFilter.class);
        m.put(ERODE, com.jhlabs.image.ErodeFilter.class);
        m.put(EXPOSURE, com.jhlabs.image.ExposureFilter.class);
        m.put(FBM, com.jhlabs.image.FBMFilter.class);
        m.put(FADE, com.jhlabs.image.FadeFilter.class);
        m.put(FEEDBACK, com.jhlabs.image.FeedbackFilter.class);
        m.put(FIELDWARP, com.jhlabs.image.FieldWarpFilter.class);
        m.put(FILL, com.jhlabs.image.FillFilter.class);
        m.put(FITINTO, org.vaadin.imagefilter.filters.FitIntoFilter.class);
        m.put(FLARE, com.jhlabs.image.FlareFilter.class);
        m.put(FLIP, com.jhlabs.image.FlipFilter.class);
        m.put(FLUSH3D, com.jhlabs.image.Flush3DFilter.class);
        m.put(FOURCOLOR, com.jhlabs.image.FourColorFilter.class);
        m.put(GAIN, com.jhlabs.image.GainFilter.class);
        m.put(GAMMA, com.jhlabs.image.GammaFilter.class);
        m.put(GAUSSIAN, com.jhlabs.image.GaussianFilter.class);
        m.put(GLINT, com.jhlabs.image.GlintFilter.class);
        m.put(GLOW, com.jhlabs.image.GlowFilter.class);
        m.put(GRADIENT, com.jhlabs.image.GradientFilter.class);
        m.put(GRADIENTWIPE, com.jhlabs.image.GradientWipeFilter.class);
        m.put(GRAY, com.jhlabs.image.GrayFilter.class);
        m.put(GRAYSCALE, com.jhlabs.image.GrayscaleFilter.class);
        m.put(HSBADJUST, com.jhlabs.image.HSBAdjustFilter.class);
        m.put(HALFTONE, com.jhlabs.image.HalftoneFilter.class);
        m.put(INTERPOLATE, com.jhlabs.image.InterpolateFilter.class);
        m.put(INVERTALPHA, com.jhlabs.image.InvertAlphaFilter.class);
        m.put(INVERT, com.jhlabs.image.InvertFilter.class);
        m.put(JAVALNF, com.jhlabs.image.JavaLnFFilter.class);
        m.put(KALEIDOSCOPE, com.jhlabs.image.KaleidoscopeFilter.class);
        m.put(KEY, com.jhlabs.image.KeyFilter.class);
        m.put(LENSBLUR, com.jhlabs.image.LensBlurFilter.class);
        m.put(LEVELS, com.jhlabs.image.LevelsFilter.class);
        m.put(LIFE, com.jhlabs.image.LifeFilter.class);
        m.put(LIGHT, com.jhlabs.image.LightFilter.class);
        m.put(LOOKUP, com.jhlabs.image.LookupFilter.class);
        m.put(MAPCOLORS, com.jhlabs.image.MapColorsFilter.class);
        m.put(MAP, com.jhlabs.image.MapFilter.class);
        m.put(MARBLE, com.jhlabs.image.MarbleFilter.class);
        m.put(MARBLETEX, com.jhlabs.image.MarbleTexFilter.class);
        m.put(MASK, com.jhlabs.image.MaskFilter.class);
        m.put(MAXIMUM, com.jhlabs.image.MaximumFilter.class);
        m.put(MEDIAN, com.jhlabs.image.MedianFilter.class);
        m.put(MINIMUM, com.jhlabs.image.MinimumFilter.class);
        m.put(MIRROR, com.jhlabs.image.MirrorFilter.class);
        m.put(MOTIONBLUR, com.jhlabs.image.MotionBlurFilter.class);
        m.put(NOISE, com.jhlabs.image.NoiseFilter.class);
        m.put(OFFSET, com.jhlabs.image.OffsetFilter.class);
        m.put(OIL, com.jhlabs.image.OilFilter.class);
        m.put(OPACITY, com.jhlabs.image.OpacityFilter.class);
        m.put(OUTLINE, com.jhlabs.image.OutlineFilter.class);
        m.put(PERSPECTIVE, com.jhlabs.image.PerspectiveFilter.class);
        m.put(PINCH, com.jhlabs.image.PinchFilter.class);
        m.put(PLASMA, com.jhlabs.image.PlasmaFilter.class);
        m.put(POINT, com.jhlabs.image.PointFilter.class);
        m.put(POINTILLIZE, com.jhlabs.image.PointillizeFilter.class);
        m.put(POLAR, com.jhlabs.image.PolarFilter.class);
        m.put(POSTERIZE, com.jhlabs.image.PosterizeFilter.class);
        m.put(QUANTIZE, com.jhlabs.image.QuantizeFilter.class);
        m.put(QUILT, com.jhlabs.image.QuiltFilter.class);
        m.put(RGBADJUST, com.jhlabs.image.RGBAdjustFilter.class);
        m.put(RAYS, com.jhlabs.image.RaysFilter.class);
        m.put(REDUCENOISE, com.jhlabs.image.ReduceNoiseFilter.class);
        m.put(RENDERTEXT, com.jhlabs.image.RenderTextFilter.class);
        m.put(RESCALE, com.jhlabs.image.RescaleFilter.class);
        m.put(RIPPLE, com.jhlabs.image.RippleFilter.class);
        m.put(ROTATE, com.jhlabs.image.RotateFilter.class);
        m.put(ROUNDEDCORNERS,
                org.vaadin.imagefilter.filters.RoundedCornersFilter.class);
        m.put(SCALE, com.jhlabs.image.ScaleFilter.class);
        m.put(SHADE, com.jhlabs.image.ShadeFilter.class);
        m.put(SHADOW, com.jhlabs.image.ShadowFilter.class);
        m.put(SHAPE, com.jhlabs.image.ShapeFilter.class);
        m.put(SHARPEN, com.jhlabs.image.SharpenFilter.class);
        m.put(SHATTER, com.jhlabs.image.ShatterFilter.class);
        m.put(SHEAR, com.jhlabs.image.ShearFilter.class);
        m.put(SKELETON, com.jhlabs.image.SkeletonFilter.class);
        m.put(SMARTBLUR, com.jhlabs.image.SmartBlurFilter.class);
        m.put(SMEAR, com.jhlabs.image.SmearFilter.class);
        m.put(SOLARIZE, com.jhlabs.image.SolarizeFilter.class);
        m.put(SPARKLE, com.jhlabs.image.SparkleFilter.class);
        m.put(SPHERE, com.jhlabs.image.SphereFilter.class);
        m.put(STAMP, com.jhlabs.image.StampFilter.class);
        m.put(SWIM, com.jhlabs.image.SwimFilter.class);
        m.put(TEXTURE, com.jhlabs.image.TextureFilter.class);
        m.put(THRESHOLD, com.jhlabs.image.ThresholdFilter.class);
        m.put(TILEIMAGE, com.jhlabs.image.TileImageFilter.class);
        m.put(TWIRL, com.jhlabs.image.TwirlFilter.class);
        m.put(UNSHARP, com.jhlabs.image.UnsharpFilter.class);
        m.put(VARIABLEBLUR, com.jhlabs.image.VariableBlurFilter.class);
        m.put(WARP, com.jhlabs.image.WarpFilter.class);
        m.put(WATER, com.jhlabs.image.WaterFilter.class);
        m.put(WEAVE, com.jhlabs.image.WeaveFilter.class);
        m.put(WOOD, com.jhlabs.image.WoodFilter.class);
        ALL_OPERATIONS_MAP = java.util.Collections.unmodifiableMap(m);
        ALL_OPERATIONS = java.util.Collections.unmodifiableSet(m.keySet());
    }

}