package org.vaadin.imagefilter;

import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;

import org.vaadin.imagefilter.utils.ImageUtils;

import com.vaadin.Application;
import com.vaadin.terminal.Resource;
import com.vaadin.terminal.StreamResource;
import com.vaadin.ui.Embedded;

public class Image extends Embedded {

    private static final long serialVersionUID = 1802095020860573587L;

    private static final String DEFAULT_IMAGE_FORMAT = "png";

    private String name;
    private URL originalUrl;
    private InputStream originalInput;
    private byte[] originalImageData;
    private boolean cacheOriginal = false;
    private BufferedImage originalImage;
    private BufferedImage currentImage;
    public String format = DEFAULT_IMAGE_FORMAT;
    private List<Operation> operations = new ArrayList<Operation>();

    /**
     * Image operation interface.
     *
     * Implement this interface to create new currentImage manipulation
     * operation. A new {@link BufferedImage} instance must be returned.
     *
     * @author Sami Ekblad
     *
     */
    public interface Operation {

        public BufferedImage apply(BufferedImage image);

        public String getName();
    }

    /**
     * Create a non-cached Image without loading any actual image data.
     *
     * @see #load(URL)
     * @see #load(String)
     * @see #load(byte[])
     * @see #load(InputStream)
     * @see #Image(boolean)
     */
    public Image() {
        this(false);
    }

    /**
     * Create an Image without loading any actual image data.
     *
     * @param cacheOriginal
     *            If the currentImage is cached the {@link #revert()} operation
     *            is fast, but it consumes more memory.
     *
     * @see #load(URL)
     * @see #load(String)
     * @see #load(InputStream)
     * @see #load(byte[])
     */

    public Image(boolean cacheOriginal) {
        this.cacheOriginal = cacheOriginal;
    }

    /**
     * Create a new Image from an URL. Original image data is not cached.
     *
     * @param url
     *            Image URL
     */
    public Image(String url) {
        load(url);
    }

    /**
     * Create a new Image from an URL.
     *
     * @param url
     * @param cacheOriginal
     *            If the currentImage is cached the {@link #revert()} operation
     *            is fast, but it consumes more memory.
     */
    public Image(String url, boolean cacheOriginal) {
        this.cacheOriginal = cacheOriginal;
        load(url);
    }

    /**
     * Create a new Image from an URL. Original image data is not cached.
     *
     * @param url
     *            Image URL
     */
    public Image(URL url) {
        load(url);
    }

    /**
     * Create a new Image from an URL.
     *
     * @param url
     * @param cacheOriginal
     *            If the currentImage is cached the {@link #revert()} operation
     *            is fast, but it consumes more memory.
     */
    public Image(URL url, boolean cacheOriginal) {
        this.cacheOriginal = cacheOriginal;
        load(url);
    }

    /**
     * Create a new Image from a {@link InputStream}. Original image data is not
     * cached.
     *
     * @param url
     *            Image input stream.
     */
    public Image(InputStream input) {
        load(input);
    }

    /**
     * Create a new Image from a {@link InputStream}. Original image data is not
     * cached.
     *
     * @param url
     *            Image input stream.
     * @param cacheOriginal
     */
    public Image(InputStream input, boolean cacheOriginal) {
        this.cacheOriginal = cacheOriginal;
        load(input);
    }

    /**
     * Create new image from given data.
     *
     * @param originalImageData
     * @param cacheOriginal
     */
    public Image(byte[] imageData, boolean cacheOriginal) {
        this.cacheOriginal = cacheOriginal;
        load(imageData);
    }

    /**
     * Load an currentImage from URL.
     *
     * @param originalUrl
     * @see ImageIO
     */
    public void load(String url) {
        try {
            load(isURL(url) ? new URL(url) : null);
        } catch (MalformedURLException e) {
        }
    }

    /**
     * Load an currentImage from URL.
     *
     * @param originalUrl
     * @see ImageIO
     */
    public void load(URL url) {
        boolean differs = url != null && !url.equals(originalUrl);
        originalUrl = url;
        originalInput = null;
        originalImageData = null;
        revert(differs);
    }

    /**
     * Create an currentImage from an {@link InputStream}. Supports all formats
     * supported by {@link ImageIO#read(InputStream)}
     *
     * @param originalImageData
     */
    public void load(InputStream imageInput) {
        originalUrl = null;
        originalInput = imageInput;
        originalImageData = null;
        revert(true);
    }

    /**
     * Create an currentImage from pre-loaded data. Supports all formats
     * supported by {@link ImageIO#read(InputStream)}
     *
     * @param originalImageData
     */
    public void load(byte[] imageData) {
        originalUrl = null;
        originalInput = null;
        originalImageData = imageData;
        revert(true);
    }

    /**
     * Revert back to original image.
     *
     */
    public void revert() {
        revert(false);
    }

    /**
     * Revert back to original image and force reload the data from the original
     * URL or byte array.
     *
     * @param forceLoad
     */
    public void revert(boolean forceLoad) {
        currentImage = null;

        // Fast revert: If not forced reload and cache image is available.
        if (!forceLoad && cacheOriginal && originalImage != null) {
            updateClientResource();
            return;
        }

        // Reload the data from original source.
        try {
            if (originalUrl != null) {
                originalImage = ImageIO.read(originalUrl);
            } else if (originalInput != null) {
                originalImage = ImageIO.read(originalInput);
            } else if (originalImageData != null) {
                originalImage = ImageIO.read(new ByteArrayInputStream(
                        originalImageData));
            }
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        name = originalUrl != null ? originalUrl.getFile() : "" + ".edited."
                + getExtensionForFormat(format);
        updateClientResource();
    }

    /**
     * Get suitable file extension for a fiven image format.
     *
     * @param format
     * @return
     */
    protected String getExtensionForFormat(String format) {
        return format;
    }

    private boolean isURL(String text) {
        try {
            new URL(text);
            return true;
        } catch (MalformedURLException ignored) {
        }
        return false;
    }

    /**
     * Vaadin {@link Resource} implementation for presenting the image in the
     * browser.
     *
     * @author Sami Ekblad
     *
     */
    public class ImageStream implements StreamResource.StreamSource,
            Serializable {

        private static final long serialVersionUID = 1L;

        public InputStream getStream() {
            return ImageUtils.createStreamSource(getImage(), format);
        }
    }

    @Override
    public void attach() {
        super.attach();
        updateClientResource();
    }

    /**
     * Updates the client resource.
     *
     * This should be called when the image content changes.
     *
     */
    protected void updateClientResource() {
        Application a = getApplication();
        if (a == null) {
            return;
        }

        // Use varying name to avoid cache problems
        StreamResource res = new StreamResource(new ImageStream(), System
                .currentTimeMillis()
                + name, a);
        setSource(res);
    }

    /**
     * Applies current image operation stack to the original image.
     *
     * Note: this reverts the image first by calling {@link #revert()}
     *
     */
    public void applyOperations() {
        revert();
        BufferedImage img = getImage();
        for (Iterator<Operation> i = operations.iterator(); i.hasNext();) {
            img = i.next().apply(img);
        }
        updateClientResource();
        currentImage = img;
    }

    /**
     * Get image data as {@link InputStream}.
     *
     * @return
     */
    public InputStream getInputStream() {
        return new ImageStream().getStream();
    }

    /**
     * Get the current image.
     *
     * If original image is cached, this always returns a second copy of the
     * image.
     *
     * @see #getOriginalImage()
     */
    public BufferedImage getImage() {
        if (currentImage == null) {
            if (cacheOriginal) {
                // Create a working copy
                currentImage = ImageUtils.createBufferedImage(originalImage
                        .getWidth(), originalImage.getHeight());
                Graphics2D g = currentImage.createGraphics();
                g.drawImage(originalImage, 0, 0, null);
                g.dispose();
            } else {
                // No caching of original image
                currentImage = originalImage;
            }
        }
        return currentImage;
    }

    /**
     * Is original image data cached.
     *
     * @return
     */
    public boolean isCacheOriginal() {
        return cacheOriginal;
    }

    /**
     * Get the original image.
     *
     * @return
     */
    public BufferedImage getOriginalImage() {
        return originalImage;
    }

    /**
     * Check if image has been successfully loaded.
     *
     * @return
     */
    public boolean isLoaded() {
        return originalImage != null && originalImage.getWidth() > 0;
    }

    /**
     * Add an image operation to the end of the stack.
     *
     * Changes to operation stack are not reflected to the image before
     * {@link #applyOperations()} is called.
     *
     * @param operation
     */
    public void addOperation(Operation operation) {
        operations.add(operation);
    }

    /**
     * Get the image operation stack. You can modify the stack directly.
     *
     * Changes to operation stack are not reflected to the image before
     * {@link #applyOperations()} is called.
     *
     *
     * @return
     */
    public List<Operation> getOperationStack() {
        return operations;
    }

    /**
     * Remove an operation from the stack.
     *
     * Changes are not reflected to the image before {@link #applyOperations()}
     * is called.
     *
     * @param operation
     */
    public void removeOperation(Operation operation) {
        operations.remove(operation);
    }

    /**
     * Move the operation by given number of steps. Negative number moves it
     * towards the start of the list and positive towards the end. For example
     * moveOperation(op,-1) moves the operation one step towards the start of
     * the list.
     *
     * Changes are not reflected to the image before {@link #applyOperations()}
     * is called.
     *
     * @param operation
     */
    public void moveOperation(Operation operation, int amount) {
        int i = operations.indexOf(operation);
        if (amount != 0 && i >= 0 && i + amount >= 0
                && i + amount < operations.size()) {
            Operation prev;
            prev = operations.get(i + amount);
            operations.set(i + amount, operation);
            operations.set(i, prev);
        }
    }

}
