package org.vaadin.imagefilter;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

import org.vaadin.imagefilter.Image.Operation;

import com.jhlabs.image.AbstractBufferedImageOp;
import com.vaadin.Application;
import com.vaadin.data.Item;
import com.vaadin.data.Property;
import com.vaadin.data.Property.ValueChangeEvent;
import com.vaadin.data.util.BeanItem;
import com.vaadin.event.Action;
import com.vaadin.event.ShortcutAction;
import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.Alignment;
import com.vaadin.ui.Button;
import com.vaadin.ui.Component;
import com.vaadin.ui.CustomField;
import com.vaadin.ui.DefaultFieldFactory;
import com.vaadin.ui.Field;
import com.vaadin.ui.Form;
import com.vaadin.ui.FormFieldFactory;
import com.vaadin.ui.HorizontalLayout;
import com.vaadin.ui.Label;
import com.vaadin.ui.ListSelect;
import com.vaadin.ui.Panel;
import com.vaadin.ui.TextField;
import com.vaadin.ui.VerticalLayout;
import com.vaadin.ui.Window;
import com.vaadin.ui.Button.ClickEvent;

public class ImageSampleApplication extends Application {

    private static final long serialVersionUID = -5679070806324266251L;

    private static final String DEFAULT_IMAGE_URL = "default";
    private static final String DEFAULT_IMAGE = "itmill_oldmill_office.png";
    private static final String LOGO_IMAGE = "vaadin-symbol-beige.png";

    private static final String WATERMARK_OP = "Add Watermark";
    private TextField url;
    private Button load;
    private Image image;
    private Panel origTitle;
    private Panel imageTitle;
    private ListSelect availableOperations;
    private ListSelect selectedOperations;
    private Form operationForm;
    private Button removeBtn;
    private Button upBtn;
    private Button downBtn;
    private Button applyButton;
    private Image origImage;

    private static BufferedImage vaadinlogo;
    static {
        try {
            vaadinlogo = ImageIO.read(ImageSampleApplication.class
                    .getResource(LOGO_IMAGE));
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private boolean disableUpdates;

    @Override
    public void init() {

        VerticalLayout baseLayout = new VerticalLayout();
        baseLayout.setMargin(true);
        baseLayout.setSpacing(true);
        // Application main window
        Window w = new Window("Image Sample", baseLayout);
        addWindow(w);

        // Layout for URL & load button
        HorizontalLayout urlLayout = new HorizontalLayout();
        urlLayout.setSpacing(true);
        w.addComponent(urlLayout);

        // TextField for image URL
        urlLayout.addComponent(new Label("Enter image url: "));
        url = new TextField(null, DEFAULT_IMAGE_URL);
        url.setColumns(40);
        urlLayout.addComponent(url);

        // Button and listener for loading the image from URL
        load = new Button("Load", new Button.ClickListener() {

            private static final long serialVersionUID = 1L;

            public void buttonClick(ClickEvent event) {
                // Because image data is cached this does not actually load
                // anything unless changed
                String u = "" + url.getValue();
                if (DEFAULT_IMAGE_URL.equals(u)) {
                    origImage.load(getDefaultImageUrl());
                    image.load(getDefaultImageUrl());
                } else if (u != null && u.startsWith("http://")) {
                    origImage.load(u);
                    image.load(u);
                }
            }
        });
        urlLayout.addComponent(load);
        urlLayout.setComponentAlignment(load, Alignment.BOTTOM_LEFT);

        // List available operations
        final List<String> opers = new ArrayList<String>();

        // All filters
        for (String f : FilterOperation.ALL_OPERATIONS) {
            opers.add(f);
        }

        // Load a logo image

        opers.add(WATERMARK_OP);

        // Select component for picking up operations
        availableOperations = new ListSelect("Add Operation:");
        availableOperations.setWidth("170px");
        availableOperations.setHeight("100%");
        for (String o : opers) {
            availableOperations.addItem(o);
        }
        availableOperations.setNullSelectionAllowed(false);
        availableOperations.setImmediate(true);
        availableOperations.addListener(new Property.ValueChangeListener() {

            private static final long serialVersionUID = 1L;

            public void valueChange(ValueChangeEvent event) {
                if (availableOperations.getValue() != null) {
                    String opName = (String) availableOperations.getValue();
                    if (opName != null) {
                        Image.Operation op = null;
                        if (WATERMARK_OP.equals(opName)) {
                            op = new DrawLogoOperation();
                        } else {
                            op = FilterOperation.getByName(opName);
                        }

                        if (op == null) {
                            return;
                        }

                        image.addOperation(op);
                        updateSelectedOperations();
                        selectedOperations.setValue(op);
                        selectedOperations.focus();
                        applyOperations();
                    }
                }
            }
        });

        // Select component for operation stack
        selectedOperations = new ListSelect("Selected Operations:");
        selectedOperations.setNullSelectionAllowed(false);
        selectedOperations.setWidth("170px");
        selectedOperations.setHeight("100%");
        selectedOperations.setImmediate(true);
        selectedOperations.addListener(new Property.ValueChangeListener() {

            private static final long serialVersionUID = 1L;

            public void valueChange(ValueChangeEvent event) {
                if (selectedOperations.getValue() != null) {
                    Image.Operation op = (Image.Operation) selectedOperations
                            .getValue();
                    if (op instanceof FilterOperation) {
                        operationForm
                                .setItemDataSource(new BeanItem<AbstractBufferedImageOp>(
                                        ((FilterOperation) op).getFilter()));
                    } else {
                        operationForm
                                .setItemDataSource(new BeanItem<Image.Operation>(
                                        op));
                    }

                    for (Object pid : operationForm.getItemPropertyIds()) {
                        Field f = operationForm.getField(pid);
                        ((AbstractComponent) f).setImmediate(true);
                        if (f.isReadOnly()) {
                            f.setVisible(false);
                        }
                    }
                } else {
                    operationForm.setItemDataSource(null);
                }
                updateOperationFormButtons();
            }
        });

        // Form for operation properties
        operationForm = new Form();
        operationForm.setWriteThrough(true);
        operationForm.setReadThrough(true);
        FormFieldFactory ff = new FilterFieldFactory();
        operationForm.setFormFieldFactory(ff);

        // Layout for form buttons
        HorizontalLayout operationButtons = new HorizontalLayout();
        operationButtons.addComponent(upBtn = new Button("up",
                new Button.ClickListener() {
                    private static final long serialVersionUID = 1L;

                    public void buttonClick(ClickEvent event) {
                        image
                                .moveOperation(
                                        (Image.Operation) selectedOperations
                                                .getValue(), -1);
                        updateSelectedOperations();
                        applyOperations();
                    }
                }));
        operationButtons.addComponent(downBtn = new Button("down",
                new Button.ClickListener() {
                    private static final long serialVersionUID = 1L;

                    public void buttonClick(ClickEvent event) {
                        image
                                .moveOperation(
                                        (Image.Operation) selectedOperations
                                                .getValue(), 1);
                        updateSelectedOperations();
                        applyOperations();
                    }
                }));
        operationButtons.addComponent(removeBtn = new Button("remove",
                new Button.ClickListener() {
                    private static final long serialVersionUID = 1L;

                    public void buttonClick(ClickEvent event) {
                        removeOperation((Image.Operation) selectedOperations
                                .getValue());
                    }
                }));

        HorizontalLayout l = new HorizontalLayout();
        l.setSpacing(true);
        w.addComponent(l);

        // Layout for operation selector
        HorizontalLayout operLayout = new HorizontalLayout();
        operLayout.setMargin(false);
        operLayout.setSpacing(true);

        Panel p = new Panel("Image Operations");
        p.addComponent(operLayout);
        l.addComponent(p);
        operLayout.addComponent(availableOperations);

        VerticalLayout selLayout = new VerticalLayout();
        operLayout.addComponent(selLayout);
        selLayout.addComponent(selectedOperations);
        selLayout.addComponent(operationButtons);

        p.addComponent(operationForm);

        applyButton = new Button("Apply", new Button.ClickListener() {

            private static final long serialVersionUID = 1L;

            public void buttonClick(ClickEvent event) {
                applyOperations();
            }
        });
        p.addComponent(applyButton);

        p.addActionHandler(new Action.Handler() {

            private static final long serialVersionUID = 1L;

            private Action[] acts = new Action[] {
                    new ShortcutAction("Apply", ShortcutAction.KeyCode.ENTER,
                            null),
                    new ShortcutAction("Remove", ShortcutAction.KeyCode.DELETE,
                            null) };

            public void handleAction(Action action, Object sender, Object target) {
                if (acts[0] == action) {
                    applyOperations();
                } else if (acts[1] == action) {
                    removeOperation((Operation) selectedOperations.getValue());
                }
            }

            public Action[] getActions(Object target, Object sender) {
                return acts;
            }
        });

        ((VerticalLayout) p.getContent()).setComponentAlignment(applyButton,
                Alignment.TOP_RIGHT);

        // A Label for title
        origTitle = new Panel("Original Image:");
        l.addComponent(origTitle);

        // A Label for title
        imageTitle = new Panel("Result:");
        l.addComponent(imageTitle);

        // Original Image
        origImage = new Image(getDefaultImageUrl(), true);
        origTitle.addComponent(origImage);

        // Manipulated cached image
        image = new Image(getDefaultImageUrl(), true);
        imageTitle.addComponent(image);

        // Put some operation as example (using filter defaults)
        disableUpdates = true;
        try {
            availableOperations.setValue(FilterOperation.EDGE);
            availableOperations.setValue(FilterOperation.GRAYSCALE);
            availableOperations.setValue(FilterOperation.FITINTO);
            availableOperations.setValue(WATERMARK_OP);
            availableOperations.setValue(FilterOperation.ROUNDEDCORNERS);
            operationForm.getField("cornerRadius").setValue(9);
            availableOperations.setValue(FilterOperation.ROTATE);
            operationForm.getField("angle").setValue(0.2);
            availableOperations.setValue(FilterOperation.DROPSHADOW);
            availableOperations.setValue(null); //Deselect
        } finally {
            disableUpdates = false;
        }
        applyOperations();

    }

    protected void removeOperation(Image.Operation o) {
        if (o == null) {
            return;
        }
        int i = image.getOperationStack().indexOf(o);
        if (i >= 0 && i < image.getOperationStack().size()) {
            image.removeOperation(o);
            updateSelectedOperations();
            applyOperations();
            if (i < image.getOperationStack().size()) {
                selectedOperations.setValue(image.getOperationStack().get(i));
            }
        }
    }

    protected void applyOperations() {
        if (!disableUpdates) {
            image.revert();
            image.applyOperations();
        }
    }

    protected void updateOperationFormButtons() {
        Object operation = selectedOperations.getValue();
        List<Image.Operation> opers = image.getOperationStack();
        boolean selected = operationForm.getItemDataSource() != null;

        applyButton.setEnabled(opers != null && opers.size() > 0);
        upBtn.setEnabled(selected && operation != opers.get(0));
        downBtn
                .setEnabled(selected
                        && operation != opers.get(opers.size() - 1));
        removeBtn.setEnabled(selected);
    }

    protected void updateSelectedOperations() {
        List<Image.Operation> opers = image.getOperationStack();
        Object selected = selectedOperations.getValue();
        selectedOperations.removeAllItems();
        for (Image.Operation o : opers) {
            selectedOperations.addItem(o);
            selectedOperations.setItemCaption(o, o.getName());
        }
        selectedOperations.setValue(selected);
    }

    private URL getDefaultImageUrl() {
        return getClass().getResource(DEFAULT_IMAGE);
    }

    public class FilterFieldFactory extends DefaultFieldFactory implements
            FormFieldFactory {

        private static final long serialVersionUID = 1L;

        @Override
        public Field createField(Item item, Object propertyId,
                Component uiContext) {
            Property p = item.getItemProperty(propertyId);
            Class<?> t = p.getType();
            if (Point2D.class.isAssignableFrom(t)) {
                PointField f = new PointField();
                f.setCaption(createCaptionByPropertyId(propertyId));
                return f;
            }
            if (String.class.equals(t) || Boolean.class.equals(t)
                    || Integer.class.equals(t) || Float.class.equals(t)
                    || Double.class.equals(t) || Byte.class.equals(t)) {
                Field f = super.createField(item, propertyId, uiContext);
                return f;
            }
            return null;
        }
    }

    public static class PointField extends CustomField {

        private static final long serialVersionUID = 1L;
        private TextField x;
        private TextField y;
        private Point2D point;
        private boolean disabled;

        public PointField() {
            HorizontalLayout layout = new HorizontalLayout();
            layout.setSpacing(true);

            layout.addComponent(new Label("X: "));
            layout.addComponent(x = new TextField());
            layout.setComponentAlignment(x, Alignment.MIDDLE_RIGHT);
            x.setColumns(3);
            layout.addComponent(new Label("Y: "));
            layout.addComponent(y = new TextField());
            layout.setComponentAlignment(x, Alignment.MIDDLE_RIGHT);
            y.setColumns(3);

            Property.ValueChangeListener l = new Property.ValueChangeListener() {

                private static final long serialVersionUID = 1L;

                public void valueChange(Property.ValueChangeEvent event) {

                    if (disabled) {
                        return;
                    }

                    Property p = event.getProperty();
                    Object v = p.getValue();
                    float f = 0;
                    if (v instanceof String) {
                        f = Float.parseFloat((String) v);
                    } else if (v instanceof Float) {
                        f = (Float) v;
                    } else {
                        return;
                    }

                    if (p == x) {
                        setValue(new Point2D.Float(f, (float) point.getY()));
                    } else {
                        setValue(new Point2D.Float((float) point.getX(), f));
                    }
                }
            };

            x.addListener(l);
            y.addListener(l);

            x.setImmediate(true);
            y.setImmediate(true);

            setCompositionRoot(layout);
        }

        @Override
        protected void setInternalValue(Object newValue) {
            super.setInternalValue(newValue);
            if (newValue instanceof Point2D) {
                point = (Point2D) newValue;
                disabled = true;
                try {
                    x.setValue(point.getX());
                    y.setValue(point.getY());
                } finally {
                    disabled = false;
                }
            }
        }

        @Override
        public Class<?> getType() {
            return Point.class;
        }

    }

    /**
     * An image operation that draws a Vaadin logo at given X/Y position.
     *
     * @author Sami Ekblad
     *
     */
    public class DrawLogoOperation extends AbstractDrawOperation {

        private int y = -5;
        private int x = -5;
        private int w = 160;
        private int h = 160;
        private float alpha = 0.5f;

        private java.awt.Image logo = null;

        @Override
        public void draw(Graphics2D g, BufferedImage srcImage) {
            g.drawImage(srcImage, 0, 0, null); // draw original

            if (logo == null) {
                logo = vaadinlogo.getScaledInstance(w, h,
                        BufferedImage.SCALE_SMOOTH);
            }
            g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getAlpha()));
            g.drawImage(logo, getX(), getY(), null); // draw the logo
        }

        @Override
        public String getName() {
            return WATERMARK_OP;
        }

        public void setY(int y) {
            this.y = y;
        }

        public int getY() {
            return y;
        }

        public void setX(int x) {
            this.x = x;
        }

        public int getX() {
            return x;
        }

        public void setWidth(int w) {
            this.w = w > 0 ? w : 16;
            logo = null;
        }

        public int getWidth() {
            return w;
        }

        public void setHeight(int h) {
            this.h = h > 0 ? h : 16;
            logo = null;
        }

        public int getHeight() {
            return h;
        }

        public void setAlpha(float alpha) {
            this.alpha = alpha;
        }

        public float getAlpha() {
            return alpha;
        }

    };

}
