package org.vaadin.imagefilter.utils;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.JarURLConnection;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

public class ClassUtils {

    public static Set<Class<?>> getClassesForPackage(String pckgname,
            boolean includeInnerClasses) throws ClassNotFoundException {
        return getClassesForPackage(pckgname, null, includeInnerClasses);
    }

    public static Set<Class<?>> getClassesForPackage(String pckgname)
            throws ClassNotFoundException {
        return getClassesForPackage(pckgname, null, false);
    }

    public static Set<Class<?>> getClassesForPackage(String pckgname,
            ClassLoader loader, boolean includeInnerClasses)
            throws ClassNotFoundException {
        // This will hold a list of directories matching the pckgname.
        // There may be more than one if a package is split over multiple
        // jars/paths
        Set<Class<?>> classes = new LinkedHashSet<Class<?>>();
        ArrayList<File> directories = new ArrayList<File>();
        try {
            ClassLoader cld = loader != null ? loader : Thread.currentThread()
                    .getContextClassLoader();
            if (cld == null) {
                throw new ClassNotFoundException("Can't get class loader.");
            }
            // Ask for all resources for the path
            String pkgPath = pckgname.replace('.', '/');
            Enumeration<URL> resources = cld.getResources(pkgPath);
            while (resources.hasMoreElements()) {
                URL res = resources.nextElement();
                if (res.getProtocol().equalsIgnoreCase("jar")) {
                    JarURLConnection conn = (JarURLConnection) res
                            .openConnection();
                    JarFile jar = conn.getJarFile();
                    for (JarEntry e : Collections.list(jar.entries())) {

                        if (e.getName().startsWith(pkgPath)
                                && e.getName().endsWith(".class")) {
                            String className = e.getName().replace("/", ".")
                                    .substring(0, e.getName().length() - 6);

                            // Innerclass hanling
                            if (className.contains("$") && !includeInnerClasses) {
                                continue;
                            }

                            try {
                                classes.add(loader != null ? loader
                                        .loadClass(className) : Class
                                        .forName(className));
                            } catch (NoClassDefFoundError ignored) {
                            } catch (UnsatisfiedLinkError ignored) {
                            }
                        }
                    }
                } else {
                    directories.add(new File(URLDecoder.decode(res.getPath(),
                            "UTF-8")));
                }
            }
        } catch (NullPointerException x) {
            throw new ClassNotFoundException(pckgname
                    + " does not appear to be "
                    + "a valid package (Null pointer exception)");
        } catch (UnsupportedEncodingException encex) {
            throw new ClassNotFoundException(pckgname
                    + " does not appear to be "
                    + "a valid package (Unsupported encoding)");
        } catch (IOException ioex) {
            throw new ClassNotFoundException(
                    "IOException was thrown when trying "
                            + "to get all resources for " + pckgname);
        }

        // For every directory identified capture all the .class files
        for (File directory : directories) {
            if (directory.exists()) {
                // Get the list of the files contained in the package
                String[] files = directory.list();
                for (String file : files) {
                    // we are only interested in .class files
                    if (file.endsWith(".class")) {
                        // removes the .class extension
                        classes.add(Class.forName(pckgname + '.'
                                + file.substring(0, file.length() - 6)));
                    }
                }
            } else {
                throw new ClassNotFoundException(pckgname + " ("
                        + directory.getPath()
                        + ") does not appear to be a valid package");
            }
        }
        return classes;
    }

    public static List<Class<?>> getClassessOfInterface(String thePackage,
            Class<?> theInterface) throws ClassNotFoundException {
        List<Class<?>> classList = new ArrayList<Class<?>>();
        for (Class<?> discovered : getClassesForPackage(thePackage)) {
            if (Arrays.asList(discovered.getInterfaces())
                    .contains(theInterface)) {
                classList.add(discovered);
            }
        }

        return classList;
    }
}
