package org.vaadin.imagefilter.utils;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.ConvolveOp;
import java.awt.image.Kernel;
import java.awt.image.RescaleOp;
import java.awt.image.WritableRaster;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import javax.imageio.ImageIO;

public class ImageUtils {

    public static InputStream createStreamSource(final BufferedImage img,
            final String formatName) {
        try {
            /* TODO: I'd like to stream data more directly */
            if (img == null) {
                return null;
            }
            ByteArrayOutputStream out2 = new ByteArrayOutputStream();
            ImageIO.write(img, formatName, out2);
            return new ByteArrayInputStream(out2.toByteArray());
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    public static BufferedImage roundCorners(BufferedImage i, int cornerRadius) {
        BufferedImage t = new BufferedImage(i.getWidth(), i.getHeight(),
                BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = t.createGraphics();
        applyDefaultRenderingHints(g);
        g.setColor(Color.WHITE);
        g.fillRoundRect(0, 0, i.getWidth(), i.getHeight(), cornerRadius * 2,
                cornerRadius * 2);
        g.setComposite(AlphaComposite.SrcIn);
        g.drawImage(i, 0, 0, null);
        g.dispose();
        return t;
    }

    public static BufferedImage shadow(BufferedImage image, int direction,
            int distance) {

        int x = distance;
        int y = distance;
        int w = image.getWidth() + distance * 2;
        int h = image.getHeight() + distance * 2;

        int ix = direction < 337.5 && direction > 202.5 ? 0 : distance;
        ix += direction > 22.5 && direction < 157.5 ? distance : 0;
        ix = distance * 2 - ix;

        int iy = direction > 67.5 && direction < 337.5 ? distance : 0;
        iy += direction > 112.5 && direction < 247.5 ? distance : 0;
        iy = distance * 2 - iy;

        // Create mask
        BufferedImage mask = new BufferedImage(w, h,
                BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = mask.createGraphics();
        applyDefaultRenderingHints(g);
        g.drawImage(image, x, y, null);
        g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_IN, 0.6f));
        g.setColor(Color.BLACK);
        g.fillRect(0, 0, mask.getWidth(), mask.getHeight());
        g.dispose();

        // Create shadow
        BufferedImage shadow = new BufferedImage(w, h,
                BufferedImage.TYPE_INT_ARGB);
        shadow = getLinearBlurOp(distance).filter(mask, shadow);

        // Compose
        BufferedImage shadowedImage = new BufferedImage(w, h,
                BufferedImage.TYPE_INT_ARGB);
        g = shadowedImage.createGraphics();
        applyDefaultRenderingHints(g);
        g.drawImage(shadow, 0, 0, null);
        g.drawImage(image, ix, iy, null);
        g.dispose();
        return shadowedImage;
    }

    public static ConvolveOp getGaussianBlurOp(int radius, boolean horizontal) {
        if (radius < 1) {
            throw new IllegalArgumentException("Radius must be >= 1");
        }

        int size = radius * 2 + 1;
        float[] data = new float[size];

        float sigma = radius / 3.0f;
        float twoSigmaSquare = 2.0f * sigma * sigma;
        float sigmaRoot = (float) Math.sqrt(twoSigmaSquare * Math.PI);
        float total = 0.0f;

        for (int i = -radius; i <= radius; i++) {
            float distance = i * i;
            int index = i + radius;
            data[index] = (float) Math.exp(-distance / twoSigmaSquare)
                    / sigmaRoot;
            total += data[index];
        }

        for (int i = 0; i < data.length; i++) {
            data[i] /= total;
        }

        Kernel kernel = null;
        if (horizontal) {
            kernel = new Kernel(size, 1, data);
        } else {
            kernel = new Kernel(1, size, data);
        }
        return new ConvolveOp(kernel, ConvolveOp.EDGE_NO_OP, null);
    }

    public static ConvolveOp getLinearBlurOp(int size) {
        float[] data = new float[size * size];
        float value = 1.0f / (size * size);
        for (int i = 0; i < data.length; i++) {
            data[i] = value;
        }
        return new ConvolveOp(new Kernel(size, size, data));
    }

    public static BufferedImage rotate(BufferedImage image, int degrees) {
        int w = image.getWidth();
        int h = image.getHeight();

        // Enough space for the image
        int nd = (int) Math.ceil(Math.sqrt(w * w + h * h)) + 2;

        BufferedImage t = new BufferedImage(nd, nd, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = t.createGraphics();
        applyDefaultRenderingHints(g);
        g.rotate(Math.toRadians(degrees), nd / 2, nd / 2);
        g.translate(nd / 2 - w / 2, nd / 2 - h / 2);

        g.drawImage(image, 0, 0, null);
        g.dispose();

        // Crop the invisible parts
        return cropTransparent(t);
    }

    public static BufferedImage cropTransparent(BufferedImage t) {
        // Find the bounding box
        WritableRaster r = t.getRaster();
        int minx = -1;
        int miny = -1;
        int maxx = r.getWidth();
        int maxy = r.getHeight();
        double[] pv = new double[4];
        int x0 = 0;
        int x1 = r.getWidth();
        int y0 = 0;
        int y1 = r.getHeight();

        // min y
        boolean contentFound = false;
        for (int y = y0; y < y1 && !contentFound; y++) {
            for (int x = x0; x < x1; x++) {
                r.getPixel(x, y, pv);
                if (pv[3] != 0) {
                    contentFound = true;
                    miny = y;
                    break;
                }
            }
        }

        // max y
        contentFound = false;
        for (int y = y1 - 1; y > 0 && !contentFound; y--) {
            for (int x = x0; x < x1; x++) {
                r.getPixel(x, y, pv);
                if (pv[3] != 0) {
                    contentFound = true;
                    maxy = y + 1;
                    break;
                }
            }
        }

        // min x
        contentFound = false;
        for (int x = x0; x < x1 && !contentFound; x++) {
            for (int y = y0; y < y1; y++) {
                r.getPixel(x, y, pv);
                if (pv[3] != 0) {
                    contentFound = true;
                    minx = x;
                    break;
                }
            }
        }

        // max x
        contentFound = false;
        for (int x = x1 - 1; x > x0 && !contentFound; x--) {
            for (int y = y0; y < y1; y++) {
                r.getPixel(x, y, pv);
                if (pv[3] != 0) {
                    contentFound = true;
                    maxx = x + 1;
                    break;
                }
            }
        }

        minx = minx < x0 ? x0 : minx;
        miny = miny < y0 ? y0 : miny;
        maxx = maxx > x1 ? x1 : maxx;
        maxy = maxy > y1 ? y1 : maxy;

        int nw = maxx - minx;
        int nh = maxy - miny;
        return crop(t, minx, miny, nw, nh);
    }

    public static BufferedImage crop(BufferedImage image, int x, int y, int w,
            int h) {
        int nw = w < image.getWidth() ? w : image.getWidth();
        int nh = h < image.getHeight() ? h : image.getHeight();
        BufferedImage t = new BufferedImage(nw, nh, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = t.createGraphics();
        applyDefaultRenderingHints(g);
        int x2 = x + w;
        x2 = x2 > image.getWidth() ? image.getWidth() : x2;
        int y2 = y + h;
        y2 = y2 > image.getHeight() ? image.getHeight() : y2;
        g.drawImage(image, 0, 0, nw, nh, x, y, x2, y2, null);
        g.dispose();
        return t;
    }

    public static BufferedImage scale(BufferedImage image, int w, int h) {
        BufferedImage t = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = t.createGraphics();
        applyDefaultRenderingHints(g);
        g.drawImage(image, 0, 0, w, h, 0, 0, image.getWidth(), image
                .getHeight(), null);
        g.dispose();
        return t;
    }

    public static BufferedImage flipHorizontal(BufferedImage image) {
        BufferedImage t = new BufferedImage(image.getWidth(),
                image.getHeight(), BufferedImage.TYPE_INT_ARGB);
        AffineTransform tx = AffineTransform.getScaleInstance(-1, 1);
        tx.translate(-image.getWidth(null), 0);
        AffineTransformOp op = new AffineTransformOp(tx,
                AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
        t = op.filter(image, t);
        return t;
    }

    public static BufferedImage flipVertical(BufferedImage image) {
        BufferedImage t = new BufferedImage(image.getWidth(),
                image.getHeight(), BufferedImage.TYPE_INT_ARGB);
        AffineTransform tx = AffineTransform.getScaleInstance(1, -1);
        tx.translate(0, -image.getHeight(null));
        AffineTransformOp op = new AffineTransformOp(tx,
                AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
        t = op.filter(image, t);
        return t;
    }

    public static BufferedImage fit(BufferedImage image, int w, int h,
            boolean cropToFit) {

        // Calculate scale
        double x_scale = (double) w / (double) image.getWidth();
        double y_scale = (double) h / (double) image.getHeight();
        double scale = x_scale < y_scale ? x_scale : y_scale;
        int nw = (int) (image.getWidth() * scale);
        int nh = (int) (image.getHeight() * scale);

        // Center if not crop
        int nx = cropToFit ? 0 : (w - nw) / 2;
        int ny = cropToFit ? 0 : (h - nh) / 2;

        BufferedImage t = new BufferedImage(cropToFit ? nw : w, cropToFit ? nh
                : h, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = t.createGraphics();
        applyDefaultRenderingHints(g);
        g.drawImage(image, nx, ny, nx + nw, ny + nh, 0, 0, image.getWidth(),
                image.getHeight(), null);
        g.dispose();
        return t;
    }

    public static BufferedImage brignessAndContrast(BufferedImage image,
            float brightness, float contrast) {
        RescaleOp op = new RescaleOp(new float[] { 0.9f, 0.9f, 0.9f },
                new float[] { 0f }, null);
        BufferedImage t = new BufferedImage(image.getWidth(),
                image.getHeight(), BufferedImage.TYPE_INT_ARGB);
        op.filter(image, t);
        return t;
    }

    /**
     * Create new buffered image.
     *
     * @param w
     * @param h
     * @return
     */
    public static BufferedImage createBufferedImage(int w, int h) {
        return new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
    }

    public static void applyDefaultRenderingHints(Graphics2D g) {
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
        g.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
                RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
                RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
    }

}
